<?php
namespace Ourgold\Enums\Enums;

use ArrayIterator;
use Countable;
use IteratorAggregate;
use ReflectionClass;
use ReflectionException;

abstract class Enum implements Countable, IteratorAggregate
{
    /**
     * Кеш рефлексии класса.
     *
     * @var ReflectionClass
     */
    protected static $reflectionCache = [];

    /**
     * Возвращает рефлексию класса.
     *
     * @return ReflectionClass
     * @throws ReflectionException
     */
    protected static function getReflection(): ReflectionClass
    {
        $class = static::class;
        static::$reflectionCache[$class] = static::$reflectionCache[$class] ?? new ReflectionClass($class);

        return static::$reflectionCache[$class];
    }

    /**
     * Получить все константы класса.
     *
     * @return array
     * @throws ReflectionException
     */
    protected static function getConstants(): array
    {
        return self::getReflection()->getConstants();
    }

    /**
     * Получить все значения констант.
     *
     * @return array
     * @throws ReflectionException
     */
    public static function getValues(): array
    {
        return array_values(static::getConstants());
    }

    /**
     * Получить ключ константы по значению.
     *
     * @param mixed $value
     * @return string
     * @throws ReflectionException
     */
    public static function getKey($value): string
    {
        return array_search($value, static::getConstants(), true);
    }

    /**
     * Получить дружественное имя константы по ключу.
     *
     * @param string $key
     * @return string
     */
    protected static function getFriendlyKeyName(string $key): string
    {
        if (ctype_upper(preg_replace('/[^a-zA-Z]/', '', $key))) {
            $key = strtolower($key);
        }

        return ucfirst(str_replace('_', ' ', snake_case($key)));
    }

    /**
     * Получить массив локализованных описаний констант.
     *
     * @return array
     */
    protected static function getLocalizedDescriptions(): array
    {
        return [];
    }

    /**
     * Получить описание константы по значению.
     *
     * @param mixed $value
     * @return string
     * @throws ReflectionException
     */
    public static function getDescription($value): string
    {
        return static::getLocalizedDescriptions()[$value]
            ?? static::getFriendlyKeyName(static::getKey($value));
    }

    /**
     * Получить форматированные константы для select.
     *
     * @return array<array-key, string>
     * @throws ReflectionException
     */
    public static function asSelectArray(): array
    {
        $array = static::getConstants();
        $selectArray = [];

        foreach ($array as $value) {
            $selectArray[$value] = static::getDescription($value);
        }

        return $selectArray;
    }

    /**
     * Массив элементов.
     *
     * @param bool $keysInLowerCase
     * @return array
     * @throws ReflectionException
     */
    public static function toArray(bool $keysInLowerCase = false): array
    {
        return ($keysInLowerCase)
            ? array_change_key_case(static::getConstants())
            : static::getConstants();
    }

    /**
     * Итератор для элементов.
     *
     * @return ArrayIterator
     * @throws ReflectionException
     */
    public function getIterator(): ArrayIterator
    {
        return new ArrayIterator(self::getConstants());
    }

    /**
     * Кол-во элементов.
     *
     * @return int
     * @throws ReflectionException
     */
    public function count(): int
    {
        return count(self::getConstants());
    }
}